#!/usr/bin/env python
'''
engine
Created by Seria at 23/11/2018 2:36 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from ..law import Law

def Engine(config=None, device=None, available_gpus='', gpu_mem_fraction=1,
                        if_conserve=True, least_mem=2048):
    if config is None:
        param = {'device': device, 'available_gpus': available_gpus, 'gpu_mem_fraction': gpu_mem_fraction,
                      'if_conserve': if_conserve, 'least_mem': least_mem}
    else:
        config['available_gpus'] = config.get('available_gpus', available_gpus)
        config['gpu_mem_fraction'] = config.get('gpu_mem_fraction', gpu_mem_fraction)
        config['if_conserve'] = config.get('if_conserve', if_conserve)
        config['least_mem'] = config.get('least_mem', least_mem)
        param = config

    core = Law.CORE.upper()
    if core == 'TENSORFLOW':
        from .engine_tf import EngineTF
        return EngineTF(param)
    else:
        raise ValueError('NEBULAE ERROR ⨷ %s has not been a supported core.' % core)


#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 3:38 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

from math import ceil
import os

from .fuel_tank import FuelTank

class FuelDepot(object):
    def __init__(self):
        self.dataset = {} # contains all fuel tanks
        self.fetcher = {} # contains batch fetcher of all datasets
        self.modifiable_keys = ['name', 'batch_size', 'height', 'width', 'channel',
                                'resol_ratio', 'spatial_aug', 'p_sa', 'temporal_aug', 'p_ta']

    def loadFuel(self, config=None, name=None, batch_size=None, if_shuffle=True,
                 data_path=None, data_key=None, is_compressed=True,
                 height=0, width=0, channel=3, frame=0, resol_ratio=1,
                 spatial_aug='', p_sa=(0), theta_sa=(0),
                 temporal_aug='', p_ta=(0), theta_ta=(0)):
        if config is None:
            config = {}
            config['name'] = name
            config['batch_size'] = batch_size
            config['data_path'] = data_path
            config['data_key'] = data_key
            config['is_compressed'] = is_compressed
            config['if_shuffle'] = if_shuffle
            config['height'] = height
            config['width'] = width
            config['channel'] = channel
            config['resol_ratio'] = resol_ratio
            config['frame'] = frame
            config['spatial_aug'] = spatial_aug
            config['p_sa'] = p_sa
            config['theta_sa'] = theta_sa
            config['temporal_aug'] = temporal_aug
            config['p_ta'] = p_ta
            config['theta_ta'] = theta_ta
        else:
            config['if_shuffle'] = config.get('if_shuffle', if_shuffle)
            config['is_compressed'] = config.get('is_compressed', is_compressed)
            config['height'] = config.get('height', height)
            config['width'] = config.get('width', width)
            config['channel'] = config.get('channel', channel)
            config['resol_ratio'] = config.get('resol_ratio', resol_ratio)
            config['frame'] = config.get('frame', frame)
            config['spatial_aug'] = config.get('spatial_aug', spatial_aug)
            config['p_sa'] = config.get('p_sa', p_sa)
            config['theta_sa'] = config.get('theta_sa', theta_sa)
            config['temporal_aug'] = config.get('temporal_aug', temporal_aug)
            config['p_ta'] = config.get('p_ta', p_ta)
            config['theta_ta'] = config.get('theta_ta', theta_ta)
        if config['name'] in self.dataset:
            raise Exception('%s is already mounted.' % name)
        if not os.path.exists(config['data_path']):
            raise ValueError('There is no such file or directory named %s.' % config['data_path'])
        self.dataset[config['name']] = FuelTank(config)
        self.fetcher[config['name']] = self.dataset[config['name']]._fetchBatches()

    def unloadFuel(self, name):
        name_err = Exception('%s is not an mounted fuel tank.' % name)
        if name not in self.dataset.keys():
            raise name_err
        else:
            self.fetcher.pop(name)
            self.dataset.pop(name)

    def nextBatch(self, name):
        return self.fetcher[name].__next__()

    def currentEpoch(self, name):
        return abs(self.dataset[name].epoch)+1

    def milesPerEpoch(self, name):
        return ceil(self.dataset[name].nsample/self.dataset[name].param['batch_size'])

    def tankVolume(self, name):
        return self.dataset[name].nsample

    def editProperty(self, dataname, config=None, **kwargs):
        flag_rename = False
        if config:
            kwargs = config
        for key in kwargs:
            if key not in self.modifiable_keys:
                print('%s is not a modifiable parameter or has not been defined.'%key)
            elif key == 'name':
                flag_rename = True
                self.dataset[dataname].param[key] = kwargs[key]
            else:
                self.dataset[dataname].param[key] = kwargs[key]
        if flag_rename:
            self.dataset[kwargs['name']] = self.dataset[dataname]
            self.dataset.pop(dataname)
            self.fetcher[kwargs['name']] = self.fetcher[dataname]
            self.fetcher.pop(dataname)


#!/usr/bin/env python
'''
space_craft_tf
Created by Seria at 03/02/2019 11:25 AM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import tensorflow as tf
import numpy as np
from .component import Pod


class SpaceCraftTF(object):
    def __init__(self, engine, blueprint, scope=''):
        assert '/' not in scope
        if scope:
            self.scope = scope + '/'
        else:
            self.scope = scope
        self.mile = tf.Variable(0, trainable=False)
        self.valid_dtypes = ['uint8', 'uint16', 'uint32', 'int8', 'int16', 'int32', 'int64',
                             'float16', 'float32', 'float64', 'str', 'bool']
        self.operands = {'+': tf.add, '-': tf.subtract, '*': tf.multiply, '@': tf.matmul, '&': tf.concat}
        self.operands_counter = {'+': 0, '-': 0, '*': 0, '@': 0, '&': 0}
        self.layout = {}
        self.engine = engine
        self.blueprint = blueprint

    def fuelLine(self, name, shape, dtype, default=None):
        if dtype not in self.valid_dtypes:
            raise Exception('NEBULAE ERROR ⨷ %s is not a valid data type.' % dtype)
        if not default is None:
            self.layout[name] = tf.placeholder_with_default(tf.constant(default, dtype=dtype),
                                                            shape, 'FL/DEFAULT_' + name)
        else:
            self.layout[name] = tf.placeholder(tf.as_dtype(dtype), shape, 'FL/' + name)

    def _getHull(self, component):
        if isinstance(component, (tf.Tensor, tf.SparseTensor, tf.Variable)):
            tshape = component.get_shape()
            if tshape.dims is None or tshape.ndims == 0:
                return ' 1    '
            shape = tshape[1:].as_list()
        elif isinstance(component, np.ndarray):
            shape = list(component.shape)
        else:
            shape = [1]
        len_wo_bs = len(shape) - 1
        hull = ' '
        for dim in range(len_wo_bs):
            hull += '%-4d x ' % shape[dim]
        hull += '%-5d' % shape[-1]
        return hull

    def _assemble(self, left_comp, len_prefix, assemblage=None, sub_scope=''):
        left_symbol = left_comp.symbol
        left_msg = left_comp.message
        if sub_scope:
            sub_prefix = sub_scope + '/'
        else:
            sub_prefix = sub_scope
        # this is an atomic component
        if left_symbol == '>':
            for comp in left_comp.component:
                assemblage = self.assemble(comp, assemblage=assemblage,
                                           sub_scope=sub_scope, is_external=False)
            return assemblage
        elif left_symbol in ['+', '-', '*', '@', '&']:
            operator = None
            self.operands_counter[left_symbol] += 1
            # TODO: if a pod is extended to more than two comps, assemblage_name needs modification
            assemblage_name = left_comp.name + '_' + str(self.operands_counter[left_symbol])
            assemblage_list = []
            hull_list = []
            init_assemblage = assemblage
            for comp in left_comp.component:
                assemblage = self.assemble(comp, assemblage=init_assemblage,
                                           sub_scope=sub_scope, is_external=False)
                assemblage_list.append(assemblage)
                hull = self._getHull(self.layout[assemblage])
                hull_list.append(hull)
                if operator is None:
                    operator = self.layout[assemblage]
                else:
                    if left_symbol == '&':
                        operator = self.operands[left_symbol]([operator, self.layout[assemblage]],
                                                              axis=-1, name=assemblage_name)
                    else:
                        operator = self.operands[left_symbol](operator,
                                                              self.layout[assemblage],
                                                              name=assemblage_name)
            self.layout[sub_prefix + assemblage_name] = operator
            self.blueprint._combineNodes(assemblage_list, hull_list,
                                            sub_prefix + assemblage_name, left_symbol)
            return sub_prefix + assemblage_name
        else:
            node_name = sub_prefix + left_comp.name
            if node_name in self.layout.keys():  # Reuse or Redefine an existent node
                assert len(left_msg) == 1  # the only way to reuse is invoking DUPLICATE with single input
                hull = self._getHull(left_msg[0])
                self.blueprint._drawNode(left_msg[0].op.name[len_prefix:], node_name, hull)
                return sub_prefix + left_comp.name
            if assemblage is None:
                self.layout[node_name] = left_comp.component(name=left_comp.name)
            else:
                if left_symbol == 'penalty':
                    self.layout[node_name] = left_comp.component(name=left_comp.name)
                else:  # message is .input
                    self.layout[node_name] = left_comp.component(name=left_comp.name,
                                                                 input=self.layout[assemblage])
                    # draw layout sheet
                    hull = self._getHull(self.layout[assemblage])
                    self.blueprint._drawNode(assemblage, node_name, hull)
            for msg in left_msg:
                # draw layout sheet
                if isinstance(msg, (tf.Tensor, tf.SparseTensor, tf.Variable)):
                    hull = self._getHull(msg)
                    if msg.name.startswith('FL/'):  # if the input is a fuel line
                        self.blueprint._drawNode(msg.op.name[3:], node_name, hull, init=True)
                    else:
                        self.blueprint._drawNode(msg.op.name[len_prefix:], node_name, hull)
            return sub_prefix + left_comp.name

    def assemble(self, left_comp, assemblage=None, sub_scope='', is_external=True,
                 reuse=None):
        if not isinstance(left_comp, Pod):
            raise TypeError('NEBULAE ERROR ⨷ only the members from Component can be assembled.')
        if self.scope[:3] == 'FL/':
            raise ValueError('NEBULAE ERROR ⨷ FL is the reserved words for fuel lines in spacecraft. '
                             + 'Please change your global scope name.')
        len_prefix = len(self.scope)
        if not is_external:
            return self._assemble(left_comp, len_prefix, assemblage=assemblage, sub_scope=sub_scope)
        else:
            # append sub scope
            if sub_scope:
                scope = self.scope + sub_scope + '/'
            else:
                scope = self.scope
            # whether to reuse variable name
            if reuse is None:
                reuse = tf.AUTO_REUSE
            with tf.variable_scope(scope, reuse=reuse):
                return self._assemble(left_comp, len_prefix, assemblage=assemblage, sub_scope=sub_scope)


#!/usr/bin/env python
'''
component
Created by Seria at 25/11/2018 2:58 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from copy import deepcopy
from ..law import Law

class Pod:
    def __init__(self, comp, symbol, name, msg=[]):
        self.component = comp
        self.symbol = symbol
        self.message = msg
        self.name = name

    def __rshift__(self, other):
        return Pod([self, other], '>', 'CASCADE')

    def __add__(self, other):
        return Pod([self, other], '+', 'ADD')

    def __sub__(self, other):
        return Pod([self, other], '-', 'SUB')

    def __mul__(self, other):
        return Pod([self, other], '*', 'MUL')

    def __matmul__(self, other):
        return Pod([self, other], '@', 'MATMUL')

    def __and__(self, other):
        return Pod([self, other], '&', 'CONCAT')

    def __pow__(self, power, modulo=None):
        assert isinstance(power, int)
        pod = deepcopy(self)
        pod._appendSerialNo('0')
        for p in range(1, power):
            pod_temp = deepcopy(self)
            pod_temp._appendSerialNo(str(p))
            pod = pod >> pod_temp
        return pod

    def _appendSerialNo(self, serial_no):
        if isinstance(self.component, list):
            self.component[0]._appendSerialNo(serial_no)
            self.component[1]._appendSerialNo(serial_no)
        else:
            self.name += '_' + serial_no

    def show(self):
        if isinstance(self.component, list):
            self.component[0].show()
            self.component[1].show()
        else:
            print(self.name)



def Component():
    core = Law.CORE.upper()
    if core == 'TENSORFLOW':
        from .component_tf import ComponentTF
        return ComponentTF()
    else:
        raise ValueError('%s has not been a supported core.' % core)


#!/usr/bin/env python
'''
navigator
Created by Seria at 23/12/2018 11:21 AM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from ..law import Law

class Stage(object):
    START = 'START'
    END = 'END'
    ACCELERATE = 'ACCELERATE'
    FORWARD = 'FORWARD'
    PREFLIGHT = 'PREFLIGHT'
    ANALYZE = 'ANALYZE'
    BREAK_DOWN = 'BREAK_DOWN'
    stages = [START, END, ACCELERATE, FORWARD, PREFLIGHT, ANALYZE, BREAK_DOWN]
    phases = [[ACCELERATE], [FORWARD]]

    @classmethod
    def register(cls, stage, group):
        stage = stage.upper()
        if stage in cls.stages:
            raise Exception('NEBULAE ERROR ⨷ %s is an existing component in warehouse.' % stage)
        exec('Stage.%s = "%s"' % (stage, stage))
        cls.stages.append(stage)
        if group.upper() in ['ACCELERATE', 'A']:
            cls.phases[0].append(stage)
        elif group.upper() in ['FORWARD', 'F']:
            cls.phases[-1].append(stage)
        else:
            raise KeyError('NEBULAE ERROR ⨷ there is no group named %s.' % group)



def Navigator(dashboard, time_machine, fuel_depot, spacecraft):
    core = Law.CORE.upper()
    if core == 'TENSORFLOW':
        from .navigator_tf import NavigatorTF
        return NavigatorTF(dashboard, time_machine, fuel_depot, spacecraft)
    else:
        raise ValueError('NEBULAE ERROR ⨷ %s has not been a supported core.' % core)


#!/usr/bin/env python
'''
layout_sheet
Created by Seria at 02/12/2018 1:20 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

from graphviz import Digraph
import os

class BluePrint(object):

    def __init__(self, config=None, verbose=True):
        if config is None:
            self.param = {'verbose': verbose}
        else:
            self.param = config
        self.layout_sheet = Digraph(comment='The Space Craft', format='jpg')
        self.nodes = []
        self.displayed = []

    def _drawNode(self, prev_node, curr_node, edge, init=False, visible=True, if_curr_displayed = False):
        '''
        :param prev_node:
        :param curr_node:
        :param edge:
        :param init: if it is the first time that prev_node appears in canvas
        :param visible:
        :param if_curr_displayed:
        :return:
        '''
        if visible:
            visible = self.param['verbose']
        if len(self.nodes) == 0 and visible:
            print('+' + ((61 + len(edge.split('x')) * 7) * '-') + '+')
        if prev_node not in self.nodes:
            self.nodes.append(prev_node)
            if init:
                self.layout_sheet.node(prev_node, prev_node, shape='doublecircle')
        if curr_node not in self.nodes:
            self.nodes.append(curr_node)
            if_curr_displayed = True
            self.layout_sheet.node(curr_node, curr_node, shape='box')
        if prev_node != curr_node:
            self.layout_sheet.edge(prev_node, curr_node, label=edge)
            if visible and (curr_node not in self.displayed):
                print('| Assembling component\033[34m%32s\033[0m | Input%s |' % (curr_node, edge))
                print('+' + ((61 + len(edge.split('x')) * 7) * '-') + '+')
            if if_curr_displayed:
                self.displayed.append(curr_node)
        elif visible:
            print('| ReAssembling component\033[34m%30s\033[0m | Input%s |' % (curr_node, edge))
            print('+' + ((61 + len(edge.split('x')) * 7) * '-') + '+')

    def _combineNodes(self, prev_node, prev_shape, curr_node, symbol):
        self.layout_sheet.node(curr_node+symbol, symbol, shape='circle')
        self.nodes.append(curr_node+symbol)
        self.displayed.append(curr_node+symbol)
        for p in range(len(prev_node)):
            pn = prev_node[p]
            ps = prev_shape[p]
            self._drawNode(pn, curr_node+symbol, ps, visible=False)
        self._drawNode(curr_node+symbol, curr_node, '', visible=False)

    def log(self, log_path):
        gv_name = os.path.join(log_path, 'layout_sheet.gv')
        self.layout_sheet.render(gv_name, view=False)
        os.rename(gv_name + '.jpg', os.path.join(log_path, 'layout_sheet.jpg'))
        os.remove(gv_name)


#!/usr/bin/env python
'''
navigator_tf
Created by Seria at 04/02/2019 4:48 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import tensorflow as tf
import time
from copy import deepcopy
from collections import defaultdict
from .navigator import Stage

class NavigatorTF(object):
    roman_numeral = ['I', 'II', 'III', 'IV', 'V', 'VI', 'VII', 'VIII', 'IX', 'X']
    def __init__(self, dashboard, time_machine, fuel_depot, spacecraft):
        config = {'NG': spacecraft.engine.param, 'TM':time_machine.param,
                  'FD': {}, 'DB': dashboard.param, 'BP': spacecraft.blueprint.param}
        ft = fuel_depot.dataset
        for k in ft.keys():
            config['FD'][k] = ft[k].param

        self.propellants = defaultdict(list)
        self.epoch = 0
        self.mile = 0
        self.mpe = 0
        self.batch = {}
        self.destination = {}
        self.macro_stage = ''
        # start a session
        self.sess = tf.Session(config=spacecraft.engine.config_proto)
        self.time_machine = time_machine
        self.fuel_depot = fuel_depot
        self.spacecraft = spacecraft
        self.dashboard = dashboard
        self.feed(Stage.START, self.time_machine._setParams,
                            sess=self.sess, mile=self.spacecraft.mile, scope=self.spacecraft.scope, config=config)
        self.feed(Stage.START, self.time_machine.backTo)

    def feed(self, stage, propellant, position=None, condition='True', **kwargs):
        if isinstance(position, int):
            quantity = len(self.propellants[stage])
            if position < quantity:
                self.propellants[stage][position] = (propellant, condition, kwargs)
            else:
                for p in range(quantity, position):
                    self.propellants[stage].append((None, 'False'))
        self.propellants[stage].append((propellant, condition, kwargs))

    def _airDrop(self, stage, n, **kwargs):
        # kwargs.update(self.propellants[stage][n][2])
        self.propellants[stage][n] = self.propellants[stage][n][0:2] + (kwargs,)

    def _triggerStage(self, curr_stage):
        for stg in Stage.stages:
            exec('%s = "%s"' % (stg, stg))
        mile = self.mile+1
        epoch = self.epoch+1
        stage = self.macro_stage
        stats = []
        cnt = 0
        for ppl, cond, kwargs in self.propellants[curr_stage]:
            cnt += 1
            if_meet = eval(cond)
            if (ppl is None) or (not if_meet):
                stats.append(None)
                continue
            stats.append(ppl(**kwargs))
        return stats

    def runAMile(self, fuel_line, pass_by=(), destination=(), interval=1):
        assert len(pass_by)+len(destination) > 0
        start = time.time()
        fl = {}
        filled_fl = {}
        quan_preflight = len(self.propellants[Stage.PREFLIGHT])
        if quan_preflight == 0: # no propellant in Preflight stage
            for i, k in enumerate(fuel_line.keys()):
                if isinstance(fuel_line[k], str):
                    fl[self.spacecraft.layout[k]] = self.batch[fuel_line[k]]
                    filled_fl[k] = self.batch[fuel_line[k]]
                else:
                    fl[self.spacecraft.layout[k]] = fuel_line[k]
                    filled_fl[k] = fuel_line[k]
        else:
            for i, k in enumerate(fuel_line.keys()):
                filled_fl[k] = self.batch[fuel_line[k]]
            for i in range(quan_preflight):
                self._airDrop(Stage.PREFLIGHT, i, fuel_line=filled_fl)
            temp_fl = self._triggerStage(Stage.PREFLIGHT)
            for tfl in temp_fl:
                if not tfl is None:
                    for k in tfl.keys():
                        fl[self.spacecraft.layout[k]] = tfl[k]
                    break

        pb = [self.spacecraft.layout[k] for k in pass_by]
        dst = [self.spacecraft.layout[k] for k in destination]
        spot = self.sess.run(pb + dst, fl)
        spot = spot[len(pass_by):len(pass_by)+len(destination)]
        sp = {}
        filled_sp = {}
        for j in range(len(destination)):
            filled_sp[destination[j]] = spot[j]
        quan_analyzer = len(self.propellants[Stage.ANALYZE])
        if quan_analyzer == 0:  # no propellants in Analyze stage
            sp = deepcopy(filled_sp)
        else:
            for j in range(quan_analyzer):
                self._airDrop(Stage.ANALYZE, j, fuel_line=filled_fl, destination=filled_sp)
            temp_sp = self._triggerStage(Stage.ANALYZE)
            for tsp in temp_sp:
                if not tsp is None:
                    sp = tsp
                    break

        prefix = self.macro_stage + ':'
        for k in sp.keys():
            self.destination[prefix + k] = sp[k]
        self.dashboard._gauge(self.destination.items(), self.mile, self.epoch, self.mpe,
                                  time.time()-start, interval)

    def runAnEpoch(self, fuel_tank, miles=-1):
        self.mpe = self.fuel_depot.milesPerEpoch(fuel_tank)
        if miles < 0:
            miles = self.mpe
        # cores = cpu_count() // 1
        # print("Creating %d-process pool" % cores)
        # pool = Pool(cores)
        for mi in range(miles):
            self.mile = mi
            # begin = time.time()
            self.batch = self.fuel_depot.nextBatch(fuel_tank)
            # print('%.3fs' % (time.time() - begin))
            for stg in Stage.phases[-1]:
                self._triggerStage(stg)

        self.destination = {}

    def log(self, guage=True, tachograph=True, blueprint=True):
        self.dashboard._log(guage, tachograph)
        if blueprint:
            self.spacecraft.blueprint.log(self.dashboard.param['log_path'])

    def launch(self, epochs=1):
        try:
            print('+' + 35 * '-' + '+')
            print('| The Spacecraft is being launched. |')
            print('+' + 35 * '-' + '+')
            self._triggerStage(Stage.START)
            if epochs < 0 or not isinstance(epochs, int):
                raise ValueError('NEBULAE ERROR ⨷ epochs must be a non-positive integer.')
            for ep in range(epochs):
                self.epoch = ep
                for stg in Stage.phases[0]:
                    self.macro_stage = stg
                    self._triggerStage(stg)
            self._triggerStage(Stage.END)
            print('+' + 44 * '-' + '+')
            print('| The Spacecraft has arrived at destination. |')
            print('+' + 44 * '-' + '+')

        except BaseException as e:
            if Stage.BREAK_DOWN in self.propellants:
                self._triggerStage(Stage.BREAK_DOWN)
            else:
                raise e


# !/usr/bin/env python
'''
space_craft
Created by Seria at 23/11/2018 10:31 AM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========

'''
# -*- coding:utf-8 -*-
from ..law import Law

def SpaceCraft(engine, blueprint, scope=''):
    core = Law.CORE.upper()
    if core == 'TENSORFLOW':
        from .spacecraft_tf import SpaceCraftTF
        return SpaceCraftTF(engine, blueprint, scope)
    else:
        raise ValueError('%s has not been a supported core.' %core)


#!/usr/bin/env python
'''
garage
Created by Seria at 03/01/2019 8:32 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from .spacecraft import SpaceCraft
from .component import Component
from os.path import basename

class Garage(object):
    def __init__(self):
        self.VGG_16 = self.vgg_16
        self.RESNET_V2_50 = self.resnet_v2_50
        self.RESNET_V2_101 = self.resnet_v2_101
        self.RESNET_V2_152 = self.resnet_v2_152
        self.SE_RESNET_V2_50 = self.se_resnet_v2_50
        self.SE_RESNET_V2_101 = self.se_resnet_v2_101
        self.SE_RESNET_V2_152 = self.se_resnet_v2_152

    def _residual(self, inputs, shallow, deep, stride, name):
        COMP = Component()
        input, is_train = inputs
        res_block = (COMP.CONV(name=name + '_conv_pre', input=input, kernel=[1, 1],
                              stride=[stride, stride], out_chs=shallow)
                    >> COMP.BATCH_NORM(name=name + '_bn_mid', is_train=is_train)
                    >> COMP.RELU(name=name + '_relu_mid')
                    >> COMP.CONV(name=name + '_conv_mid', kernel=[3, 3], out_chs=shallow)
                    >> COMP.BATCH_NORM(name=name + '_bn_post', is_train=is_train)
                    >> COMP.RELU(name=name + '_relu_post')
                    >> COMP.CONV(name=name + '_conv_post', kernel=[1, 1], out_chs=deep))
        return res_block

    def _bottleNeckRes(self, sc, inputs, shallow, deep, scope, name, down_sample=False, pre_activate=True):
        COMP = Component()
        input, is_train = inputs
        if pre_activate:
            input = sc.assemble(COMP.BATCH_NORM(name=name + '_bn_pre',
                                                  input=sc.layout[input], is_train=sc.layout[is_train])
                                  >> COMP.RELU(name=name + '_relu_pre'), sub_scope=scope)
        if down_sample:
            stride = 2
            idt_path = (COMP.CONV(name=name+'_conv_idt', input=sc.layout[input],
                                  kernel=[1, 1], stride=[2, 2], out_chs=deep))
        else:
            stride = 1
            if pre_activate:
                idt_path = (COMP.DUPLICATE(name=basename(input), input=sc.layout[input]))
            else:
                idt_path = (COMP.CONV(name=name + '_conv_idt', input=sc.layout[input], kernel=[1, 1], out_chs=deep))

        res_path = self._residual((sc.layout[input], sc.layout[is_train]), shallow, deep, stride, name)
        return sc.assemble(idt_path + res_path, sub_scope=scope)

    def _bottleNeckSE(self, sc, inputs, shallow, deep, scope, name, down_sample=False, pre_activate=True):
        COMP = Component()
        input, is_train = inputs
        if pre_activate:
            input = sc.assemble(COMP.BATCH_NORM(name=name + '_bn_pre',
                                                input=sc.layout[input], is_train=sc.layout[is_train])
                                >> COMP.RELU(name=name + '_relu_pre'), sub_scope=scope)
        if down_sample:
            stride = 2
            idt_path = (COMP.CONV(name=name + '_conv_idt', input=sc.layout[input],
                                  kernel=[1, 1], stride=[2, 2], out_chs=deep))
        else:
            stride = 1
            if pre_activate:
                idt_path = (COMP.DUPLICATE(name=basename(input), input=sc.layout[input]))
            else:
                idt_path = (COMP.CONV(name=name + '_conv_idt', input=sc.layout[input], kernel=[1, 1], out_chs=deep))

        res_path = sc.assemble(self._residual((sc.layout[input], sc.layout[is_train]),
                                               shallow, deep, stride, name), sub_scope=scope)
        se_path = (COMP.DUPLICATE(name=basename(res_path), input=sc.layout[res_path])
                   * self._squeezeExicitation(sc.layout[res_path], shallow//4, deep, name))

        return sc.assemble(idt_path + se_path, sub_scope=scope)

    def _squeezeExicitation(self, input, shallow, deep, name):
        COMP = Component()
        se_block = (COMP.AVG_POOL(name=name+'_se_avg_pool', input=input, padding='valid', if_global=True)
                    >> COMP.CONV(name=name+'_se_down', kernel=[1,1], out_chs=shallow)
                    >> COMP.RELU(name=name+'_se_relu')
                    >> COMP.CONV(name=name+'_se_up', kernel=[1,1], out_chs=deep)
                    >> COMP.SIGMOID(name=name+'_se_sigm'))
        return se_block

    def vgg_16(self, sc, inputs, scope='VGG_16', p_drop=0.5):
        COMP = Component()
        COMP.addComp('conv_3x3', COMP.CONV(name='conv_3x3', kernel=[3, 3]))
        conv1 = (COMP.CONV_3X3(name='conv1_1', input=sc.layout[inputs], out_chs=64)
                 >> COMP.RELU(name='relu1_1')
                 >> COMP.CONV_3X3(name='conv1_2', out_chs=64)
                 >> COMP.RELU(name='relu1_2')
                 >> COMP.MAX_POOL(name='max_pool1'))
        conv2 = ((COMP.CONV_3X3(name='conv2', out_chs=128)
                 >> COMP.RELU(name='relu2')) ** 2
                 >> COMP.MAX_POOL(name='max_pool2'))
        conv3 = ((COMP.CONV_3X3(name='conv3', out_chs=256)
                 >> COMP.RELU(name='relu3')) ** 3
                 >> COMP.MAX_POOL(name='max_pool3'))
        conv4 = ((COMP.CONV_3X3(name='conv4', out_chs=512)
                 >> COMP.RELU(name='relu4')) ** 3
                 >> COMP.MAX_POOL(name='max_pool4'))
        conv5 = ((COMP.CONV_3X3(name='conv5', out_chs=512)
                     >> COMP.RELU(name='relu5')) ** 3
                 >> COMP.MAX_POOL(name='max_pool5'))
        gconv6 = (COMP.CONV(name='gconv6', kernel=[7, 7], out_chs=4096, padding='valid')
                 >> COMP.DROPOUT(name='dropout6', p_drop=p_drop))
        gconv7 = (COMP.CONV(name='gconv7', kernel=[1, 1], out_chs=4096)
                 >> COMP.DROPOUT(name='dropout7', p_drop=p_drop))
        net = sc.assemble(conv1 >> conv2 >> conv3 >> conv4 >> conv5 >> gconv6 >> gconv7, sub_scope=scope)
        return net

    def resnet_v2_50(self, sc, input, is_train, scope='Res_50'):
        return self._resnet_v2(sc, input, is_train, scope, [3, 4, 6, 3])

    def resnet_v2_101(self, sc, input, is_train, scope='Res_101'):
        return self._resnet_v2(sc, input, is_train, scope, [3, 4, 23, 3])

    def resnet_v2_152(self, sc, input, is_train, scope='Res_152'):
        return self._resnet_v2(sc, input, is_train, scope, [3, 8, 36, 3])

    def _resnet_v2(self, sc, input, is_train, scope, blocks):
        COMP = Component()
        net = sc.assemble(COMP.CONV(name='conv0', input=sc.layout[input],
                                    kernel=[7, 7], stride=[2, 2], out_chs=64)
                 >> COMP.MAX_POOL(name='max_pool0', kernel=[3, 3]), sub_scope=scope)
        # building block 1
        net = self._bottleNeckRes(sc, [net, is_train], 64, 256, scope, pre_activate=False, name='block1_0')
        for l in range(1, blocks[0]):
            net = self._bottleNeckRes(sc, [net, is_train], 64, 256, scope, name='block1_'+str(l))
        # building block 2
        net = self._bottleNeckRes(sc, [net, is_train], 128, 512, scope, down_sample=True, name='block2_0')
        for l in range(1, blocks[1]):
            net = self._bottleNeckRes(sc, [net, is_train], 128, 512, scope, name='block2_'+str(l))
        # building block 3
        net = self._bottleNeckRes(sc, [net, is_train], 256, 1024, scope, down_sample=True, name='block3_0')
        for l in range(1, blocks[2]):
            net = self._bottleNeckRes(sc, [net, is_train], 256, 1024, scope, name='block3_' + str(l))
        # building block 4
        net = self._bottleNeckRes(sc, [net, is_train], 512, 2048, scope, down_sample=True, name='block4_0')
        for l in range(1, blocks[3]):
            net = self._bottleNeckRes(sc, [net, is_train], 512, 2048, scope, name='block4_' + str(l))
        return sc.assemble(COMP.AVG_POOL(name='avg_pool', input=sc.layout[net],
                                         kernel=[7, 7], padding='valid', if_global=True), sub_scope=scope)

    def se_resnet_v2_50(self, sc, input, is_train, scope='SE_Res_50'):
        return self._se_resnet_v2(sc, input, is_train, scope, [3, 4, 6, 3])

    def se_resnet_v2_101(self, sc, input, is_train, scope='SE_Res_101'):
        return self._se_resnet_v2(sc, input, is_train, scope, [3, 4, 23, 3])

    def se_resnet_v2_152(self, sc, input, is_train, scope='SE_Res_152'):
        return self._se_resnet_v2(sc, input, is_train, scope, [3, 8, 36, 3])

    def _se_resnet_v2(self, sc, input, is_train, scope, blocks):
        COMP = Component()
        net = sc.assemble(COMP.CONV(name='conv0', input=sc.layout[input],
                                    kernel=[7, 7], stride=[2, 2], out_chs=64)
                 >> COMP.MAX_POOL(name='max_pool0', kernel=[3, 3]), sub_scope=scope)
        # building block 1
        net = self._bottleNeckSE(sc, [net, is_train], 64, 256, scope, pre_activate=False, name='block1_0')
        for l in range(1, blocks[0]):
            net = self._bottleNeckSE(sc, [net, is_train], 64, 256, scope, name='block1_'+str(l))
        # building block 2
        net = self._bottleNeckSE(sc, [net, is_train], 128, 512, scope, down_sample=True, name='block2_0')
        for l in range(1, blocks[1]):
            net = self._bottleNeckSE(sc, [net, is_train], 128, 512, scope, name='block2_'+str(l))
        # building block 3
        net = self._bottleNeckSE(sc, [net, is_train], 256, 1024, scope, down_sample=True, name='block3_0')
        for l in range(1, blocks[2]):
            net = self._bottleNeckSE(sc, [net, is_train], 256, 1024, scope, name='block3_' + str(l))
        # building block 4
        net = self._bottleNeckSE(sc, [net, is_train], 512, 2048, scope, down_sample=True, name='block4_0')
        for l in range(1, blocks[3]):
            net = self._bottleNeckSE(sc, [net, is_train], 512, 2048, scope, name='block4_' + str(l))
        return sc.assemble(COMP.AVG_POOL(name='avg_pool', input=sc.layout[net],
                                         kernel=[7, 7], padding='valid', if_global=True), sub_scope=scope)


#!/usr/bin/env python
'''
time_machine_tf
Created by Seria at 04/02/2019 4:35 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import tensorflow as tf
from tensorflow.python.framework import graph_util
import os
import json
from ..toolkit import parseConfig, recordConfig

class TimeMachineTF(object):
    def __init__(self, param):
        '''
        Time Machine saves current states or restores saved states
        '''
        self.param = param
        self.restorer = None
        self.saver = None
        if not self.param['ckpt_path'] is None:
            if self.param['ckpt_scope'] is None:
                to_be_restored = tf.global_variables()
            else:
                to_be_restored = tf.global_variables(scope=self.param['ckpt_scope'])
            self.restorer = tf.train.Saver(to_be_restored)
        if not self.param['save_path'] is None:
            if self.param['save_scope'] is None:
                to_be_saved = tf.global_variables()
            else:
                to_be_saved = tf.global_variables(scope=self.param['save_scope'])
            self.saver = tf.train.Saver(to_be_saved, max_to_keep=1)
        if not self.param['thawed_path'] is None:
            self.time_point = tf.GraphDef()
            with open(self.param['thawed_path'], 'rb') as thawing:
                self.time_point.ParseFromString(thawing.read())

    def _setParams(self, sess, mile, scope, config):
        self.config = config
        self.sess = sess
        self.mile = mile
        if scope:
            self.scope = scope[:-1]
        else:
            self.scope = 'ckpt'
        self.sess.run(tf.global_variables_initializer())

    def backTo(self):
        if self.restorer is None:
            return
        else:
            if os.path.isfile(self.param['ckpt_path']):
                ckpt = self.param['ckpt_path']
            else:
                ckpt = tf.train.latest_checkpoint(self.param['ckpt_path'])
            self.restorer.restore(self.sess, ckpt)
            print('+' + ((10 + len(self.param['ckpt_path'])) * '-') + '+')
            print('| Back to \033[1;34m%s\033[0m |' % self.param['ckpt_path'])
            print('+' + ((10 + len(self.param['ckpt_path'])) * '-') + '+')

    def digWormHole(self):
        if self.saver is None:
            return
        else:
            if not os.path.exists(self.param['save_path']):
                os.mkdir(self.param['save_path'])
            self.saver.save(self.sess, os.path.join(self.param['save_path'], self.scope),
                            global_step=self.mile, write_meta_graph=True)
            print('| Worm Hole is located at \033[1;34m%s\033[0m |' % self.param['save_path'])

            temp_config = os.path.join(os.getcwd(), 'temp_config.json')
            if os.path.exists(temp_config):
                hyper_param = parseConfig(temp_config)
                self.config['SC'] = hyper_param
                recordConfig(os.path.join(self.param['save_path'], 'config.json'), self.config)
                os.remove(os.path.join(os.getcwd(), 'temp_config.json'))

    def thawTheMoment(self, fuel_line, moments):
        moments = [m.name for m in moments]
        return tf.import_graph_def(self.time_point, fuel_line, return_elements=moments)

    def freezeTheMoment(self, moments):
        if self.param['frozen_path'] is None:
            return
        else:
            moments = [m.op.name for m in moments]
            out_graph_def = graph_util.convert_variables_to_constants(self.sess, self.sess.graph_def, moments)
            tf.train.write_graph(out_graph_def, os.path.dirname(self.param['frozen_path']),
                                 os.path.basename(self.param['frozen_path']), as_text=False)
            print('+' + ((23 + len(self.param['frozen_path'])) * '-') + '+')
            print('| Moment is frozen in \033[1;34m%s\033[0m |' % self.param['frozen_path'])
            print('+' + ((23 + len(self.param['frozen_path'])) * '-') + '+')


#!/usr/bin/env python
'''
time_machine
Created by Seria at 23/12/2018 8:34 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from ..law import Law

def TimeMachine(config=None, ckpt_path=None, ckpt_scope=None,
                 save_path=None, save_scope=None, thawed_path=None, frozen_path=None):
    if config is None:
        param = {'ckpt_path': ckpt_path, 'ckpt_scope': ckpt_scope,
                      'save_path': save_path, 'save_scope': save_scope,
                      'thawed_path': thawed_path, 'frozen_path': frozen_path}
    else:
        config['ckpt_path'] = config.get('ckpt_path', ckpt_path)
        config['ckpt_scope'] = config.get('ckpt_scope', ckpt_scope)
        config['save_path'] = config.get('save_path', save_path)
        config['save_scope'] = config.get('save_scope', save_scope)
        config['thawed_path'] = config.get('thawed_path', thawed_path)
        config['frozen_path'] = config.get('frozen_path', frozen_path)
        param = config

    core = Law.CORE.upper()
    if core == 'TENSORFLOW':
        from .time_machine_tf import TimeMachineTF
        return TimeMachineTF(param)
    else:
        raise ValueError('NEBULAE ERROR ⨷ %s has not been a supported core.' % core)


#!/usr/bin/env python
'''
setup
Created by Seria at 04/11/2018 10:50 AM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nebulae",
    version="0.1.12",
    author="Seria",
    author_email="zzqsummerai@yeah.net",
    description="A novel and simple framework based on prevalent DL framework and other image processing libs."
                + " v0.0.12: change the quoting character of label file to |;"
                + " add fillFuel and deductFuel in toolkit;"
                + " users can store data with variable length;"
                + " reorganize the logic for assembling components.",
                # + " v0.0.11: rename LayoutSheet as BluePrint;"
                # + " change the initial parts of Space Craft and Navigator;"
                # + " merge log function in Aerolog as an interface exposed by Navigator.",
                # + " v0.1.10: make it simpler to use Dash Board module alone.",
                # + " v0.1.9: be able to remove EXIF without modifying raw images.",
                # + " v0.1.7: users can generate hdf5 as several files since generating large dataset at once is risky;"
                # + " In addition, mergeFuel function is provided for merging multiple hdf5 files;"
                # + " users can remove EXIF in images while generating data file by setting keep_exif as False.",
                # + " v0.1.6: add SE Resnets to Garage;"
                # + " read image in RGB mode of which number of channel is 3.",
                # + " v0.1.5: fix a bug would return wrong device id when looking for available gpu.",
                # + " v0.1.1: change the way to register stage;"
                # + " fix a bug in DashBoard may draw points in wrong places.",
                # + " v0.1.0: A roughly complete version. New parts, Engine, Time Machine, Dashboard and Navigator are added.",
                # + " v0.0.18: fix a bug would repeatedly append variable scope.",
                # + " v0.0.17: network layout will be saved as image instead of pdf;"
                # + " add components: RESHAPE, SLICE, CLIP;"
                # + " implement Time Machine.",
                # + " v0.0.15: fix a bug would lead to wrong implementation of ** symbol."
                # + " v0.0.14: unable to assemble DUPLICATE component without name but passing existent name is allowed."
                # + " v0.0.13: allow users to assemble DUPLICATE component without passing name;"
                # + " add RESIZE component;"
                # + " shorten automatically generated component names."
                # + " v0.0.12: fix wrong implementation on data augmentation;"
                # + " FuelGenerator will keep original image size if width or height is not given.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/",
    packages=setuptools.find_packages(),
    install_requires=
    ['graphviz',
     'h5py',
     'pillow',
     'piexif',
     'pandas'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


#!/usr/bin/env python
'''
utility
Created by Seria at 14/11/2018 8:33 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

import json
import numpy as np
import subprocess as subp
import h5py
import os

from ..law import Law

def toDenseLabel(labels, nclass, on_value=1, off_value=0):
    batch_size = len(labels)
    # initialize dense labels
    dense = off_value * np.ones((batch_size * nclass))
    indices = []
    if isinstance(labels[0], str):
        for b in range(batch_size):
            indices += [int(s) + b * nclass for s in labels[b].split(Law.CHAR_SEP)]
    else: # labels is a nested list
        for b in range(batch_size):
            indices += [l + b * nclass for l in labels[b]]
        dense[indices] = on_value
    return np.reshape(dense, (batch_size, nclass))


def getAvailabelGPU(least_mem):
    p = subp.Popen('nvidia-smi', stdout=subp.PIPE)
    gpu_id = 0  # next gpu we are about to probe
    flag_gpu = False
    max_vacancy = least_mem
    id_best = -1  # gpu having max avialable memory
    for l in p.stdout.readlines():
        line = l.decode('utf-8').split()
        if len(line) < 1:
            break
        elif len(line) < 2:
            continue
        if line[1] == str(gpu_id):
            flag_gpu = True
            continue
        if flag_gpu:
            vacancy = int(line[10].split('M')[0]) - int(line[8].split('M')[0])
            if vacancy > max_vacancy:
                max_vacancy = vacancy
                id_best = gpu_id
            gpu_id += 1
            flag_gpu = False
    return id_best

def parseConfig(config_path):
    with open(config_path, 'r') as config_file:
        return json.load(config_file)

def recordConfig(config_path, config):
    with open(config_path, 'w') as config_file:
        json.dump(config, config_file, indent=4)

def mergeFuel(src_dir, src, dst, dtype, data_key='', keep_src=True):
    data = {}
    info_keys = []
    shards = len(src)
    print('+' + (23 + 2 * shards + len(dst)) * '-' + '+')
    # read
    ending_char = '\r'
    for i, f in enumerate(src):
        hdf5 = h5py.File(os.path.join(src_dir, f), 'r')
        if i == 0:
            for key in hdf5.keys():
                info_keys.append(key)
            for key in info_keys:
                if key == Law.FRAME_KEY:
                    data[key] = 0
                else:
                    data[key] = []
        for key in info_keys:
            if key == Law.FRAME_KEY:
                frames = hdf5[key]
                data[key] = frames if frames>data[key] else data[key]
            else:
                data[key].extend(hdf5[key][()].tolist())
        hdf5.close()
        if not keep_src:
            os.remove(os.path.join(src_dir, f))

        progress = i+1
        yellow_bar = progress * '❖-'
        space_bar = (shards - progress) * '◇-'
        if progress == shards:
            ending_char = '\n'
        print('| Merging data  \033[1;35m%s\033[0m  ⊰⟦-%s%s⟧⊱ |'
              % (dst, yellow_bar, space_bar), end=ending_char)
    # write
    hdf5 = h5py.File(os.path.join(src_dir, dst), 'w')
    for key in info_keys:
        if key == data_key:  # dealing with raw data
            dt = h5py.special_dtype(vlen=dtype[key])
            hdf5.create_dataset(key, dtype=dt, data=np.array(data[key]))
        elif key == Law.FRAME_KEY:
            hdf5[key] = data[key]
        else:
            hdf5[key] = np.array(data[key]).astype(dtype[key])
    hdf5.close()
    print('+' + (23 + 2 * shards + len(dst)) * '-' + '+')

def fillFuel(src, key, data):
    hdf5 = h5py.File(src, 'a')
    hdf5[key] = data
    hdf5.close()

def deductFuel(src, key):
    hdf5 = h5py.File(src, 'a')
    del hdf5[key]
    hdf5.close()


#!/usr/bin/env python
'''
dash_board
Created by Seria at 29/12/2018 3:29 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import matplotlib.pyplot as plt
import scipy.misc as misc
import numpy as np
import time
import pandas as pd
from os import path

class DashBoard(object):
    palette = ['#F08080', '#00BFFF', '#FFFF00', '#2E8B57', '#6A5ACD', '#FFD700', '#808080']
    linestyle = ['-', '--', '-.', ':']
    def __init__(self, config=None, log_path='./aerolog', window=1, format=None):
        '''
        :param config:
        :param window: the window length of moving average
        :param format: a list of which the element is format and mode, e.g. ['3f', 'raw']
        '''
        if config is None:
            self.param = {'log_path': log_path, 'window': window, 'format': format}
        else:
            config['window'] = config.get('window', window)
            self.param = config
        self.max_epoch = 0
        self.win_mile = {}
        self.gauge_mile = {}
        self.gauge_epoch = {}
        self.trail_mile = {}
        self.trail_epoch = {}

    def _getOridinal(self, number):
        remainder = number % 10
        if remainder == 1:
            ordinal = 'st'
        elif remainder == 2:
            ordinal = 'nd'
        elif remainder == 3:
            ordinal = 'rd'
        else:
            ordinal = 'th'
        return ordinal

    def _formatAsStr(self, stage, abbr, value, global_mile=-1):
        format, mode = self.param['format'][abbr]
        format = '%-' + format
        if mode == 'raw':
            return (' %%s ➠ \033[1;36m%s\033[0m |' % format) % (abbr, value)
        elif mode == 'percent':
            return (' %%s ➠ \033[1;36m%s%%%%\033[0m |' % format) % (abbr, value*100)
        elif mode == 'image':
            if global_mile >= 0:
                misc.imsave(path.join(self.param['log_path'], '%s/%s_%d.jpg'%(stage, abbr, global_mile)))
            return ''
        else:
            raise KeyError('%s is an illegal format option.' % mode)

    def _gauge(self, items, mile, epoch, mpe, duration, interval):
        epoch += 1
        string_mile = ''
        stage = 'UNKNOWN'
        flag_display = False
        flag_epoch_end = False
        if mile % interval == 0:
            flag_display = True
        if (mile+1) % mpe == 0:
            flag_epoch_end = True
            if epoch > self.max_epoch:
                self.max_epoch = epoch
            string_epoch = ''
            cnt = 0
        if mile == 0:
            self.time = time.time()
        for name, value in items:
            # read gauge every mile
            global_mile = ((epoch-1)*mpe+mile)
            if flag_display:
                stage, abbr = name.split(':')
                string_mile += self._formatAsStr(stage, abbr, value, global_mile)
            if name not in self.win_mile.keys():
                self.win_mile[name] = np.zeros((self.param['window'],))
                self.gauge_mile[name] = []
                self.gauge_epoch[name] = [0]
                self.trail_mile[name] = []
                self.trail_epoch[name] = []
            self.win_mile[name][global_mile % self.param['window']] = value
            self.gauge_epoch[name][-1] += value
            if global_mile < self.param['window']:
                gauge = np.array(self.win_mile[name][:global_mile+1]).mean()
            else:
                gauge = np.array(self.win_mile[name]).mean()
            self.gauge_mile[name].append(gauge)
            self.trail_mile[name].append(global_mile)
            if flag_epoch_end:
                # read gauge every epoch
                stage, abbr = name.split(':')
                self.gauge_epoch[name][-1] /= mpe
                self.trail_epoch[name].append(epoch)
                indicator = self._formatAsStr(stage, abbr, self.gauge_epoch[name][-1])
                string_epoch += indicator
                if indicator != '':
                    cnt += 1
                self.gauge_epoch[name].append(0)
        if flag_display:
            ordinal = self._getOridinal(epoch)
            progress = int(mile / mpe * 10 + 0.4)
            yellow_bar = progress * ' '
            space_bar = (10 - progress) * ' '
            print('| %d%s Epoch ✇ %d Miles ⊰⟦\033[43m%s\033[0m%s⟧⊱︎ ⧲ %.3fs/mile | %s |%s     '
                  % (epoch, ordinal, mile, yellow_bar, space_bar, duration, stage, string_mile), end='\r')
        if flag_epoch_end:
            ordinal = self._getOridinal(epoch)
            mileage = str(epoch * mpe)
            display = '| %d%s Epoch ✇ %s Miles ︎⧲ %.2fs/epoch | %s |%s' \
                      % (epoch, ordinal, mileage, time.time() - self.time, stage, string_epoch)
            print('+' + (len(display) - 3 - cnt * 11) * '-' + '+' + 30 * ' ')
            print(display)
            print('+' + (len(display) - 3 - cnt * 11) * '-' + '+' + 30 * ' ')
            self.time = time.time()

    def _log(self, guage=True, tachograph=True):
        if guage:
            boards = {}
            # clustering
            for k in self.param['format'].keys():
                boards[k] = []
            for k in self.gauge_mile.keys():
                boards[k.split(':')[-1]].append(k)
            # plot
            for k in boards.keys():
                for i, b in enumerate(boards[k]):
                    plt.plot(self.trail_mile[b], self.gauge_mile[b], c=self.palette[i % 7], label=b)
                    plt.legend()
                    plt.grid(True)
                    plt.savefig(path.join(self.param['log_path'], '%s%d_%.3g_mile_%d.jpg'
                                          % (k, i, self.gauge_mile[b][-1], self.trail_mile[b][-1])))
                    plt.close()

                for i, b in enumerate(boards[k]):
                    plt.plot(self.trail_epoch[b], self.gauge_epoch[b][:-1], marker='o',
                             c=self.palette[i % 7], linestyle=self.linestyle[i % 4], label=b)
                if self.max_epoch > 0:
                    plt.legend()
                    plt.grid(True)
                    plt.savefig(path.join(self.param['log_path'], '%s_epoch_%d.jpg' % (k, self.max_epoch)))
                    plt.close()
        if tachograph:
            for k in self.gauge_mile.keys():
                stage, metric = k.split(':')
                df = pd.DataFrame(data={'A': self.trail_mile[k], 'B': self.gauge_mile[k]})
                df.to_csv(path.join(self.param['log_path'], '%s_%s_mile.csv'%(metric, stage)),
                          header=None, index=None)
                df = pd.DataFrame(data={'A': self.trail_epoch[k], 'B': self.gauge_epoch[k][:-1]})
                df.to_csv(path.join(self.param['log_path'], '%s_%s_epoch.csv' % (metric, stage)),
                          header=None, index=None)